/*****************************************************************************
 * aliases.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000-2002 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net, cpinkham@bc2va.org
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#include "config.h"

#include <sys/stat.h>

#include "gyach.h"
#include "aliases.h"
#include "commands.h"
#include "interface.h"
#include "main.h"
#include "support.h"
#include "users.h"
#include "util.h"
#include "yahoochat.h"

int  alias_count;
int  alias_selected = -1;

GtkWidget *chat_aliases;
GtkWidget *alias_menu;
GtkWidget *alias_window;

/*********************************/
/* Default aliases list */
/*********************************/
struct ALIAS {
	char *alias_name;
	char *alias_everyone;
	char *alias_user;
};

static struct ALIAS aliases[] = {
	{ "Agree",	"agrees whole-heartedly.", "agrees with $0." },
	{ "Apologize",	"apologizes.", "apologizes to $0." },
	{ "Applaud",	"applauds fervently.", "applauds $0 fervently hmmm." },
	{ "Beam",	"beams.", "beams at $0." },
	{ "Beg",	"begs like a dog.", "begs $0." },
	{ "Blush",	"blushes in embarassment.", "blushes at $0 in embarassment." },
	{ "Bow",	"bows gracefully.", "bows to $0 gracefully." },
	{ "Cheer",	"cheers enthusiastically.", "cheers $0 enthusiastically." },
	{ "Clap",	"claps briefly.", "claps for $0 briefly." },
	{ "Disagree",	"disagrees.", "disagrees with $0." },
	{ "Faint",	"faints dead away.", "faints dead away at the sight of $0's." },
	{ "Flirt",	"flirts.", "flirts with $0." },
	{ "Grin",	"grins evilly.", "grins at $0 evilly." },
	{ "Grovel",	"grovels.", "grovels at $0's feet." },
	{ "Hug",	"waves Hi to everyone.", "gives $0 a great big bear hug." },
	{ "Kiss",	"kisses.", "kisses $0." },
	{ "Laugh",	"laughs hysterically.", "laughs at $0." },
	{ "Listen",	"listens.", "listens to $0 hanging on her every word." },
	{ "Melt",	"melts.", "melts in $0's arms." },
	{ "Misgrin",	"grins mischievously.", "grins at $0 mischievously." },
	{ "Nod",	"nods solemnly.", "nods to $0 approvingly." },
	{ "Nudge",	"nudges everyone in the side.", "nudges $0 in the side." },
	{ "Pout",	"pouts.", "pouts at $0." },
	{ "Sit",	"sits down.", "sits down next to $0." },
	{ "Smile",	"smiles.", "smiles at $0." },
	{ "Stand",	"stands up.", "stands up next to $0." },
	{ "Wave",	"waves at everyone.", "waves at $0." },
	{ NULL, NULL, NULL }
};

/*
 * alias datafile fmt is "alias]everyone|individual" similar to
 * CheetaChat but support's multiple arguments per alias ($0, $1, ... $9 )
 */
void read_aliases() {
	gchar *listitem[3];
	FILE *fp;
	struct ALIAS *alias;
	char line[1024];
	char *alias_name = NULL;
	char *alias_everyone = NULL;
	char *alias_user = NULL;
	char *ptr = NULL;
	char alias_filename[1024] = "";
	char tmp_filename[1024] = "";
	int row;
	struct stat sbuf;

	DBG( 1, "read_aliases()\n" );

	gtk_clist_set_sort_column(GTK_CLIST(chat_aliases), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(chat_aliases), 1 );

	gtk_clist_freeze(GTK_CLIST(chat_aliases));

	/* remove any items already in the list */
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_aliases), 0, 0, &alias_name )) {
			break;
		}
		gtk_clist_remove( GTK_CLIST(chat_aliases), 0 );
	}

	listitem[0] = alias_name;
	listitem[1] = alias_everyone;
	listitem[2] = alias_user;
	alias_name = line;

	if ( alias_filename[ 0 ] == '\0' )
#ifdef OS_WINDOWS
		sprintf( alias_filename, "./aliases" );
#else
		sprintf( alias_filename, "%s/.gyach/aliases",
			getenv( "HOME" ));
#endif

	fp = fopen( alias_filename, "r" );

	/* if couldn't open aliases, check for emotions.dat & rename */
	if ( ! fp ) {
#ifdef OS_WINDOWS
		sprintf( tmp_filename, "./emotions.dat" );
#else
		sprintf( tmp_filename, "%s/.gyach/emotions.dat",
			getenv( "HOME" ));
#endif

		if ( ! stat( tmp_filename, &sbuf )) {
			/* file exists, so rename and open */
			rename( tmp_filename, alias_filename );
			fp = fopen( alias_filename, "r" );
		}
	}

	alias_count = 0;

	if ( fp ) {
		while( fgets( line, 1023, fp )) {
			if ( ! strchr( line, ']' ))	continue;

			while(( line[strlen(line)-1] == '\r' ) ||
				  ( line[strlen(line)-1] == '\n' )) {
				line[strlen(line)-1] = '\0';
			}

			while(( ptr = strstr( line, "%N" )) != NULL ) {
				*ptr = '$';
				*(ptr+1) = '0';
			}

			alias_count++;

			ptr = line;
			while( *ptr != ']' ) {
				*ptr = tolower( *ptr );
				ptr++;
			}

			line[0] = toupper( line[0] );
			ptr = strchr( line, ']' );
			*ptr = '\0';

			alias_everyone = ptr + 1;
			ptr++;

			ptr = strchr( alias_everyone, '|' );
			*ptr = '\0';
			alias_user = ptr + 1;

			listitem[0] = alias_name;
			row = gtk_clist_insert(GTK_CLIST(chat_aliases), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_aliases), row, 1, alias_everyone );
			gtk_clist_set_text(GTK_CLIST(chat_aliases), row, 2, alias_user );
		}
		fclose( fp );
	} else {
		/* no alias file so set defaults */
		alias = &aliases[0];
		while( alias->alias_name ) {
			listitem[0] = alias->alias_name;
			row = gtk_clist_insert(GTK_CLIST(chat_aliases), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_aliases), row, 1,
				alias->alias_everyone );
			gtk_clist_set_text(GTK_CLIST(chat_aliases), row, 2,
				alias->alias_user );
			
			alias++;
		}

		/* save our default set to the file */
		save_aliases();
	}

	gtk_clist_thaw(GTK_CLIST(chat_aliases));
}

void save_aliases() {
	gchar *alias_name;
	gchar *alias_everyone;
	gchar *alias_user;
	char alias_filename[1024] = "";
	int i = 0;
	FILE *fp;

	DBG( 1, "save_aliases()\n" );

	if ( alias_filename[ 0 ] == '\0' )
#ifdef OS_WINDOWS
		sprintf( alias_filename, "./aliases" );
#else
		sprintf( alias_filename, "%s/.gyach/aliases",
			getenv( "HOME" ));
#endif

	fp = fopen( alias_filename, "wb" );

	if ( fp ) {
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(chat_aliases), i, 0,
					&alias_name)) {
				break;
			}

			gtk_clist_get_text(GTK_CLIST(chat_aliases), i, 1, &alias_everyone );
			gtk_clist_get_text(GTK_CLIST(chat_aliases), i, 2, &alias_user );

			fprintf( fp, "%s]%s|%s\n", alias_name, alias_everyone, alias_user );
			
			i++;
		}
		fclose( fp );
	}
}

int alias_find( char *alias ) {
	gchar *alias_ptr;
	int i;
	int found = -1;

	DBG( 11, "alias_find( '%s' )\n", alias );

	i = 0;
	while( 1 ) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_aliases), i, 0, &alias_ptr ))
			break;

		if ( ! strcasecmp( alias, alias_ptr )) {
			found = i;
			break;
		}

		i++;
	}

	DBG( 12, "RETURN alias_find() == %d\n", found );

	return( found );
}

void edit_alias() {
	gchar *listitem[3];
	gchar alias_name[128] = "";
	gchar alias_everyone[128] = "";
	gchar alias_user[128] = "";
	GtkWidget *w_alias_name;
	GtkWidget *w_alias_everyone;
	GtkWidget *w_alias_user;

	listitem[0] = alias_name;
	listitem[1] = alias_everyone;
	listitem[2] = alias_user;

	DBG( 11, "edit_alias()\n" );

	if ( alias_selected > -1 ) {
		gtk_clist_get_text( GTK_CLIST(chat_aliases), alias_selected, 0,
			listitem );
		gtk_clist_get_text( GTK_CLIST(chat_aliases), alias_selected, 1,
			listitem+1 );
		gtk_clist_get_text( GTK_CLIST(chat_aliases), alias_selected, 2,
			listitem+2 );
	}

    if ( ! alias_window ) {
		alias_window = create_alias_window();
		gtk_widget_show( alias_window );
	}

	w_alias_name = lookup_widget( alias_window, "alias_name" );
	w_alias_everyone = lookup_widget( alias_window, "alias_everyone" );
	w_alias_user = lookup_widget( alias_window, "alias_user" );

	gtk_entry_set_text( GTK_ENTRY(w_alias_name), listitem[0] );
	gtk_entry_set_text( GTK_ENTRY(w_alias_everyone), listitem[1] );
	gtk_entry_set_text( GTK_ENTRY(w_alias_user), listitem[2] );
}

int check_alias( char *alias, char *args )
{
	gchar *listitem[3];
	gchar *alias_name;
	gchar *alias_everyone;
	gchar *alias_user;
	u_char  tmp[4096] = "";
	u_char  tmp2[4096] = "";
	u_char  tmp_alias[21];
	int i, x;
	int found = -1;

	DBG( 21, "check_alias( %p, '%s', '%s' )\n", sess, alias, args );

	listitem[0] = alias_name;
	listitem[1] = alias_everyone;
	listitem[2] = alias_user;

	for( i = 0; i < strlen( alias ); i++ ) {
		alias[i] = tolower( alias[i] );
	}

	i = 0;
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_aliases), i, 0, &alias_name ))
			break;

		strcpy( tmp_alias, alias );
		for( x = 0; x < strlen( tmp_alias ); x++ ) {
			tmp_alias[x] = tolower( tmp_alias[x] );
		}

		if ( ! strcmp( alias, tmp_alias )) {
			alias_name = tmp_alias;
			found = i;
			break;
		}

		i++;
	}

	if ( found >= 0 ) {
		if ( args[0] ) {
			gtk_clist_get_text(GTK_CLIST(chat_aliases), found, 2,
				&alias_user );

			strcpy( tmp2, alias_user );

			if ( strchr( tmp2, '$' )) {
				strcpy( tmp2, replace_args( tmp2, args ));
				args[0] = '\0';
			}

			alias_user = tmp2;
			if (( alias_user[0] == '/' ) ||
				( alias_user[0] == '!' )) {
				strcpy( tmp, alias_user );
				if ( args[0] ) {
					strcat( tmp, " " );
					strcat( tmp, args );
				}
				chat_command( tmp );
				DBG( 22, "RETURN check_alias() == TRUE\n" );

				return TRUE;
			}

			strcpy( tmp, YAHOO_COLOR_PURPLE );
			strcat( tmp, alias_user );

		} else {
			gtk_clist_get_text(GTK_CLIST(chat_aliases), found, 1,
				&alias_everyone );

			if (( alias_everyone[0] == '/' ) ||
				( alias_everyone[0] == '!' )) {
				chat_command( alias_everyone );
				DBG( 22, "RETURN check_alias() == TRUE\n" );

				return TRUE;
			}

			strcat( tmp, YAHOO_COLOR_PURPLE );
			strcpy( tmp, alias_everyone );
		}

		subst_escs( tmp );

		if ( my_status )
			cmd_away( "0" );

		if( send_avatar ) {
			strcat( tmp, AVATAR_START );
			strcat( tmp, send_avatar );
			strcat( tmp, AVATAR_SEND_END );
#ifndef OS_WINDOWS
			if ( show_avatars ) {
				display_avatar( ymsg_sess->user, send_avatar );
			}
#endif
		}

		ymsg9_emote( ymsg_sess, tmp );

		DBG( 22, "RETURN check_alias() == TRUE\n" );

		return( TRUE );
	} else {
		DBG( 22, "RETURN check_alias() == FALSE\n" );

		return( FALSE );
	}
}
